@@ -1,6 +1,5 @@
 from __future__ import annotations
 
-import glob
 import json
 import os
 import platform
@@ -8,27 +7,48 @@ import shutil
 import subprocess
 import sys
 import sysconfig
-from distutils import log
-from distutils.errors import (
+import logging
+import warnings
+from setuptools.errors import (
     CompileError,
-    DistutilsExecError,
-    DistutilsFileError,
-    DistutilsPlatformError,
+    ExecError,
+    FileError,
+    PlatformError,
 )
-from distutils.sysconfig import get_config_var
+from sysconfig import get_config_var
 from pathlib import Path
-from typing import Dict, Iterable, List, NamedTuple, Optional, Set, Tuple, cast
+from typing import Dict, List, Literal, NamedTuple, Optional, Set, Tuple, cast
 
-import pkg_resources
-from setuptools.command.build import build as CommandBuild  # type: ignore[import]
+from setuptools import Distribution
 from setuptools.command.build_ext import build_ext as CommandBuildExt
 from setuptools.command.build_ext import get_abi3_suffix
-from typing_extensions import Literal
+from setuptools.command.install_scripts import install_scripts as CommandInstallScripts
 
 from ._utils import format_called_process_error
 from .command import RustCommand
 from .extension import Binding, RustBin, RustExtension, Strip
-from .rustc_info import get_rust_host, get_rust_target_list, get_rustc_cfgs
+from .rustc_info import (
+    get_rust_host,
+    get_rust_version,
+    get_rustc_cfgs,
+)
+
+logger = logging.getLogger(__name__)
+
+
+try:
+    from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel
+except ImportError:  # wheel installation might be deferred in PEP 517
+    from setuptools import Command as CommandBdistWheel
+
+
+def _check_cargo_supports_crate_type_option() -> bool:
+    version = get_rust_version()
+
+    if version is None:
+        return False
+
+    return version.major > 1 or (version.major == 1 and version.minor >= 64)  # type: ignore
 
 
 class build_rust(RustCommand):
@@ -55,30 +75,27 @@ class build_rust(RustCommand):
     ]
     boolean_options = ["inplace", "debug", "release", "qbuild"]
 
+    inplace: bool = False
+    debug: bool = False
+    release: bool = False
+    qbuild: bool = False
+
     plat_name: Optional[str] = None
+    build_temp: Optional[str] = None
 
     def initialize_options(self) -> None:
         super().initialize_options()
-        self.inplace = None
-        self.debug = None
-        self.release = None
-        self.qbuild = None
-        self.build_temp = None
-        self.plat_name = None
-        self.build_number = None
         self.target = os.getenv("CARGO_BUILD_TARGET")
         self.cargo = os.getenv("CARGO", "cargo")
 
     def finalize_options(self) -> None:
         super().finalize_options()
 
-        self.data_dir = self.get_data_dir()
-
-        if self.plat_name is None:
-            self.plat_name = cast(  # type: ignore[no-any-unimported]
-                CommandBuild, self.get_finalized_command("build")
-            ).plat_name
-            assert isinstance(self.plat_name, str)
+        # Inherit settings from the `build` and `build_ext` commands
+        self.set_undefined_options(
+            "build",
+            ("plat_name", "plat_name"),
+        )
 
         # Inherit settings from the `build_ext` command
         self.set_undefined_options(
@@ -88,17 +105,11 @@ class build_rust(RustCommand):
             ("inplace", "inplace"),
         )
 
-        if self.build_number is not None and not self.build_number[:1].isdigit():
-            raise ValueError("Build tag (build-number) must start with a digit.")
-
-    def get_data_dir(self) -> str:
-        components = (
-            pkg_resources.safe_name(self.distribution.get_name()).replace("-", "_"),  # type: ignore[attr-defined]
-            pkg_resources.safe_version(self.distribution.get_version()).replace("-", "_"),  # type: ignore[attr-defined]
-        )
-        if self.build_number:
-            components += (self.build_number,)
-        return "-".join(components) + ".data"
+        if self.build_temp is not None:
+            warnings.warn(
+                "`--build-temp` argument does nothing for Rust extensions, set `CARGO_TARGET_DIR` instead.",
+                DeprecationWarning,
+            )
 
     def run_for_extension(self, ext: RustExtension) -> None:
         assert self.plat_name is not None
@@ -130,19 +141,26 @@ class build_rust(RustCommand):
     def build_extension(
         self, ext: RustExtension, forced_target_triple: Optional[str] = None
     ) -> List["_BuiltModule"]:
-
         target_triple = self._detect_rust_target(forced_target_triple)
         rustc_cfgs = get_rustc_cfgs(target_triple)
 
         env = _prepare_build_environment()
 
         if not os.path.exists(ext.path):
-            raise DistutilsFileError(
-                f"can't find Rust extension project file: {ext.path}"
+            raise FileError(
+                f"can't find manifest for Rust extension `{ext.name}` at path `{ext.path}`"
             )
 
         quiet = self.qbuild or ext.quiet
         debug = self._is_debug_build(ext)
+        use_cargo_crate_type = _check_cargo_supports_crate_type_option()
+
+        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
+        if package_id is None:
+            raise FileError(
+                f"manifest for Rust extention `{ext.name}` at path `{ext.path}` is a virtual manifest (a workspace root without a package).\n\n"
+                "If you intended to build for a workspace member, set `path` for the extension to the member's Cargo.toml file."
+            )
 
         cargo_args = self._cargo_args(
             ext=ext, target_triple=target_triple, release=not debug, quiet=quiet
@@ -161,14 +179,21 @@ class build_rust(RustCommand):
             ]
 
         else:
-            rustc_args = [
-                "--crate-type",
-                "cdylib",
-                *ext.rustc_flags,
-            ]
+            # If toolchain >= 1.64.0, use '--crate-type' option of cargo.
+            # See https://github.com/PyO3/setuptools-rust/issues/320
+            if use_cargo_crate_type:
+                rustc_args = [
+                    *ext.rustc_flags,
+                ]
+            else:
+                rustc_args = [
+                    "--crate-type",
+                    "cdylib",
+                    *ext.rustc_flags,
+                ]
 
-            # OSX requires special linker arguments
-            if rustc_cfgs.get("target_os") == "macos":
+            # Apple platforms require special linker arguments
+            if rustc_cfgs.get("target_os") in {"macos", "ios", "tvos", "watchos"}:
                 ext_basename = os.path.basename(self.get_dylib_ext_path(ext, ext.name))
                 rustc_args.extend(
                     [
@@ -188,7 +213,10 @@ class build_rust(RustCommand):
                 "wasm32",
                 "emscripten",
             ):
-                rustc_args.extend(["-C", f"link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
+                rustc_args.extend(["-C", "link-args=-sSIDE_MODULE=2 -sWASM_BIGINT"])
+
+            if use_cargo_crate_type and "--crate-type" not in cargo_args:
+                cargo_args.extend(["--crate-type", "cdylib"])
 
             command = [
                 self.cargo,
@@ -233,7 +261,7 @@ class build_rust(RustCommand):
             raise CompileError(format_called_process_error(e, include_stdout=False))
 
         except OSError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "Unable to execute 'cargo' - this package "
                 "requires Rust to be installed and cargo to be on the PATH"
             )
@@ -242,14 +270,13 @@ class build_rust(RustCommand):
         # it into the build directory as if it were produced by build_ext.
 
         dylib_paths = []
-        package_id = ext.metadata(quiet=quiet)["resolve"]["root"]
 
         if ext._uses_exec_binding():
             # Find artifact from cargo messages
             artifacts = _find_cargo_artifacts(
                 cargo_messages.splitlines(),
                 package_id=package_id,
-                kind="bin",
+                kinds={"bin"},
             )
             for name, dest in ext.target.items():
                 if not name:
@@ -262,7 +289,7 @@ class build_rust(RustCommand):
                         if Path(artifact).with_suffix("").name == name
                     )
                 except StopIteration:
-                    raise DistutilsExecError(
+                    raise ExecError(
                         f"Rust build failed; unable to locate executable '{name}'"
                     )
 
@@ -274,20 +301,18 @@ class build_rust(RustCommand):
                 dylib_paths.append(_BuiltModule(dest, artifact_path))
         else:
             # Find artifact from cargo messages
-            artifacts = tuple(
-                _find_cargo_artifacts(
-                    cargo_messages.splitlines(),
-                    package_id=package_id,
-                    kind="cdylib",
-                )
+            artifacts = _find_cargo_artifacts(
+                cargo_messages.splitlines(),
+                package_id=package_id,
+                kinds={"cdylib", "dylib"},
             )
             if len(artifacts) == 0:
-                raise DistutilsExecError(
-                    "Rust build failed; unable to find any build artifacts"
+                raise ExecError(
+                    "Rust build failed; unable to find any cdylib or dylib build artifacts"
                 )
             elif len(artifacts) > 1:
-                raise DistutilsExecError(
-                    f"Rust build failed; expected only one build artifact but found {artifacts}"
+                raise ExecError(
+                    f"Rust build failed; expected only one cdylib or dylib build artifact but found {artifacts}"
                 )
 
             artifact_path = artifacts[0]
@@ -321,33 +346,65 @@ class build_rust(RustCommand):
                 )
 
             if ext._uses_exec_binding():
-                ext_path = build_ext.get_ext_fullpath(module_name)
-                # remove extensions
-                ext_path, _, _ = _split_platform_and_extension(ext_path)
-
-                # Add expected extension
                 exe = sysconfig.get_config_var("EXE")
-                if exe is not None:
-                    ext_path += exe
 
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
                 if isinstance(ext, RustBin):
-                    executable_name = module_name
+                    # will install the rust binary into the scripts directory
+                    bin_name = module_name
                     if exe is not None:
-                        executable_name += exe
-                    scripts_dir = os.path.join(
-                        build_ext.build_lib, self.data_dir, "scripts"
+                        bin_name += exe
+
+                    install_scripts = cast(
+                        CommandInstallScripts,
+                        self.get_finalized_command("install_scripts"),
                     )
-                    os.makedirs(scripts_dir, exist_ok=True)
-                    ext_path = os.path.join(scripts_dir, executable_name)
+                    ext_path = os.path.join(install_scripts.build_dir, bin_name)
                 else:
+                    # will install the rust binary into the module directory
+                    ext_path = build_ext.get_ext_fullpath(module_name)
+
+                    # add expected extension
+                    ext_path, _, _ = _split_platform_and_extension(ext_path)
+                    if exe is not None:
+                        ext_path += exe
+
+                    # if required, also generate a console script entry point
                     ext.install_script(module_name.split(".")[-1], ext_path)
             else:
+                # will install the rust library into the module directory
                 ext_path = self.get_dylib_ext_path(ext, module_name)
-                os.makedirs(os.path.dirname(ext_path), exist_ok=True)
 
-            log.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
-            shutil.copyfile(dylib_path, ext_path)
+            os.makedirs(os.path.dirname(ext_path), exist_ok=True)
+
+            # Make filenames relative to cwd where possible, to make logs and
+            # errors below a little neater
+
+            cwd = os.getcwd()
+            if dylib_path.startswith(cwd):
+                dylib_path = os.path.relpath(dylib_path, cwd)
+            if ext_path.startswith(cwd):
+                ext_path = os.path.relpath(ext_path, cwd)
+
+            logger.info("Copying rust artifact from %s to %s", dylib_path, ext_path)
+
+            # We want to atomically replace any existing library file. We can't
+            # just copy the new library directly on top of the old one as that
+            # causes the existing library to be modified (rather the replaced).
+            # This means that any process that currently uses the shared library
+            # will see it modified and likely segfault.
+            #
+            # We first copy the file to the same directory, as `os.replace`
+            # doesn't work across file system boundaries.
+            temp_ext_path = ext_path + "~"
+            shutil.copyfile(dylib_path, temp_ext_path)
+            try:
+                os.replace(temp_ext_path, ext_path)
+            except PermissionError as e:
+                msg = f"{e}\n  hint: check permissions for {ext_path!r}"
+                if sys.platform == "win32":
+                    # On Windows, dll files are locked by the system when in use.
+                    msg += "\n  hint: the file may be in use by another Python process"
+                raise CompileError(msg)
 
             if sys.platform != "win32" and not debug_build:
                 args = []
@@ -360,7 +417,7 @@ class build_rust(RustCommand):
                     args.insert(0, "strip")
                     args.append(ext_path)
                     try:
-                        output = subprocess.check_output(args)
+                        subprocess.check_output(args)
                     except subprocess.CalledProcessError:
                         pass
 
@@ -411,17 +468,13 @@ class build_rust(RustCommand):
         return ext_path
 
     def _py_limited_api(self) -> _PyLimitedApi:
-        bdist_wheel = self.distribution.get_command_obj("bdist_wheel", create=False)
+        bdist_wheel = _get_bdist_wheel_cmd(self.distribution, create=False)
 
         if bdist_wheel is None:
             # wheel package is not installed, not building a limited-api wheel
             return False
         else:
-            from wheel.bdist_wheel import bdist_wheel as CommandBdistWheel
-
-            bdist_wheel_command = cast(CommandBdistWheel, bdist_wheel)  # type: ignore[no-any-unimported]
-            bdist_wheel_command.ensure_finalized()
-            return cast(_PyLimitedApi, bdist_wheel_command.py_limited_api)
+            return cast(_PyLimitedApi, bdist_wheel.py_limited_api)
 
     def _detect_rust_target(
         self, forced_target_triple: Optional[str] = None
@@ -467,10 +520,10 @@ class build_rust(RustCommand):
         if target_triple is not None:
             args.extend(["--target", target_triple])
 
-        if release:
-            profile = ext.get_cargo_profile()
-            if not profile:
-                args.append("--release")
+        ext_profile = ext.get_cargo_profile()
+        env_profile = os.getenv("SETUPTOOLS_RUST_CARGO_PROFILE")
+        if release and not ext_profile and not env_profile:
+            args.append("--release")
 
         if quiet:
             args.append("-q")
@@ -491,6 +544,18 @@ class build_rust(RustCommand):
         if ext.args is not None:
             args.extend(ext.args)
 
+        if env_profile:
+            if ext_profile:
+                args = [p for p in args if not p.startswith("--profile=")]
+                while True:
+                    try:
+                        index = args.index("--profile")
+                        del args[index : index + 2]
+                    except ValueError:
+                        break
+
+            args.extend(["--profile", env_profile])
+
         if ext.cargo_manifest_args is not None:
             args.extend(ext.cargo_manifest_args)
 
@@ -510,7 +575,7 @@ def create_universal2_binary(output_path: str, input_paths: List[str]) -> None:
         try:
             from fat_macho import FatWriter
         except ImportError:
-            raise DistutilsExecError(
+            raise ExecError(
                 "failed to locate `lipo` or import `fat_macho.FatWriter`. "
                 "Try installing with `pip install fat-macho` "
             )
@@ -547,9 +612,18 @@ def _replace_vendor_with_unknown(target: str) -> Optional[str]:
 def _prepare_build_environment() -> Dict[str, str]:
     """Prepares environment variables to use when executing cargo build."""
 
+    base_executable = None
+    if os.getenv("SETUPTOOLS_RUST_PEP517_USE_BASE_PYTHON"):
+        base_executable = getattr(sys, "_base_executable")
+
+    if base_executable and os.path.exists(base_executable):
+        executable = os.path.realpath(base_executable)
+    else:
+        executable = sys.executable
+
     # Make sure that if pythonXX-sys is used, it builds against the current
     # executing python interpreter.
-    bindir = os.path.dirname(sys.executable)
+    bindir = os.path.dirname(executable)
 
     env = os.environ.copy()
     env.update(
@@ -559,9 +633,9 @@ def _prepare_build_environment() -> Dict[str, str]:
             # interpreter from the path.
             "PATH": os.path.join(bindir, os.environ.get("PATH", "")),
             "PYTHON_SYS_EXECUTABLE": os.environ.get(
-                "PYTHON_SYS_EXECUTABLE", sys.executable
+                "PYTHON_SYS_EXECUTABLE", executable
             ),
-            "PYO3_PYTHON": os.environ.get("PYO3_PYTHON", sys.executable),
+            "PYO3_PYTHON": os.environ.get("PYO3_PYTHON", executable),
         }
     )
     return env
@@ -607,12 +681,12 @@ def _binding_features(
                 python_version = py_limited_api[2:]
                 features.add(f"pyo3/abi3-py{python_version}")
             elif py_limited_api:
-                features.add(f"pyo3/abi3")
+                features.add("pyo3/abi3")
         return features
     elif ext.binding is Binding.RustCPython:
         return {"cpython/python3-sys", "cpython/extension-module"}
     else:
-        raise DistutilsPlatformError(f"unknown Rust binding: '{ext.binding}'")
+        raise PlatformError(f"unknown Rust binding: '{ext.binding}'")
 
 
 _PyLimitedApi = Literal["cp37", "cp38", "cp39", "cp310", "cp311", "cp312", True, False]
@@ -659,23 +733,23 @@ def _find_cargo_artifacts(
     cargo_messages: List[str],
     *,
     package_id: str,
-    kind: str,
-) -> Iterable[str]:
+    kinds: Set[str],
+) -> List[str]:
     """Identifies cargo artifacts built for the given `package_id` from the
     provided cargo_messages.
 
-    >>> list(_find_cargo_artifacts(
+    >>> _find_cargo_artifacts(
     ...    [
     ...        '{"some_irrelevant_message": []}',
     ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
-    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
+    ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["dylib", "rlib"]},"filenames":["/file/two/baz.dylib", "/file/two/baz.rlib"]}',
     ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
     ...    ],
     ...    package_id="some_id",
-    ...    kind="cdylib",
-    ... ))
+    ...    kinds={"cdylib", "dylib"},
+    ... )
     ['/some/path/baz.so', '/file/two/baz.dylib']
-    >>> list(_find_cargo_artifacts(
+    >>> _find_cargo_artifacts(
     ...    [
     ...        '{"some_irrelevant_message": []}',
     ...        '{"reason":"compiler-artifact","package_id":"some_id","target":{"kind":["cdylib"]},"filenames":["/some/path/baz.so"]}',
@@ -683,13 +757,14 @@ def _find_cargo_artifacts(
     ...        '{"reason":"compiler-artifact","package_id":"some_other_id","target":{"kind":["cdylib"]},"filenames":["/not/this.so"]}',
     ...    ],
     ...    package_id="some_id",
-    ...    kind="rlib",
-    ... ))
+    ...    kinds={"rlib"},
+    ... )
     ['/file/two/baz.rlib']
     """
+    artifacts = []
     for message in cargo_messages:
         # only bother parsing messages that look like a match
-        if "compiler-artifact" in message and package_id in message and kind in message:
+        if "compiler-artifact" in message and package_id in message:
             parsed = json.loads(message)
             # verify the message is correct
             if (
@@ -699,8 +774,9 @@ def _find_cargo_artifacts(
                 for artifact_kind, filename in zip(
                     parsed["target"]["kind"], parsed["filenames"]
                 ):
-                    if artifact_kind == kind:
-                        yield filename
+                    if artifact_kind in kinds:
+                        artifacts.append(filename)
+    return artifacts
 
 
 def _replace_cross_target_dir(path: str, ext: RustExtension, *, quiet: bool) -> str:
@@ -714,3 +790,14 @@ def _replace_cross_target_dir(path: str, ext: RustExtension, *, quiet: bool) ->
     cross_target_dir = ext._metadata(cargo="cross", quiet=quiet)["target_directory"]
     local_target_dir = ext._metadata(cargo="cargo", quiet=quiet)["target_directory"]
     return path.replace(cross_target_dir, local_target_dir)
+
+
+def _get_bdist_wheel_cmd(  # type: ignore[no-any-unimported]
+    dist: Distribution, create: Literal[True, False] = True
+) -> Optional[CommandBdistWheel]:
+    try:
+        cmd_obj = dist.get_command_obj("bdist_wheel", create=create)
+        cmd_obj.ensure_finalized()  # type: ignore[union-attr]
+        return cast(CommandBdistWheel, cmd_obj)  # type: ignore[no-any-unimported]
+    except Exception:
+        return None
